package dev.olshevski.navigation.reimagined

import android.os.Parcelable
import androidx.compose.runtime.Stable
import androidx.lifecycle.Lifecycle
import androidx.lifecycle.ViewModelStore
import androidx.savedstate.SavedStateRegistry
import kotlinx.parcelize.Parcelize
import kotlinx.parcelize.RawValue

/**
 * A single item in a backstack with its own identity. The backstack may contain several equal
 * destinations and this wrapper class is what helps to differentiate them.
 *
 * The identity is preserved by [id] property. Every entry with unique id in the NavHost will
 * be served its own set of components such as a [Lifecycle], a [SavedStateRegistry] and
 * a [ViewModelStore]. Additionally, it will be able to remember all saveable structures
 * within composition (`rememberSaveable`).
 *
 * Always compare [NavEntry] instances by id. Even if you place the same instance into
 * the backstack twice, after a process/activity recreation it will be restored as two independent
 * instances with the same id.
 */
@Stable
@Parcelize
class NavEntry<T> internal constructor(

    /**
     * Autogenerated unique id.
     */
    val id: NavId = NavId(),

    /**
     * A destination you've passed into [navigate], [replaceLast] or other extension method.
     * Or if you've used [NavController.setNewBackstackEntries] directly this is the destination
     * you've passed into [navEntry] method.
     */
    val destination: @RawValue T
) : Parcelable {

    override fun toString(): String {
        return "NavEntry(id=$id, destination=$destination)"
    }

}

/**
 * Creates a new [NavEntry] with autogenerated unique id and the specified [destination].
 */
fun <T> navEntry(destination: T) = NavEntry(destination = destination)